import numpy as np

def binstep(x):
    return 1 if x > 0 else (0 if x == 0 else -1)
    
def linear(x):
    return x
    
def sigmoid(x):
    return 1/(1 + np.exp(-x))
    
def tanh(x):
    return np.tanh(x)
    
def relu(x):
    return x if x > 0 else 0
    
def lrelu(x):
    return x if x > 0 else 0.1 * x
    
def elu(x):
    return x if x >= 0 else np.exp(x) - 1

def swish(x):
    return x * 1/(1 + np.exp(-x)) 
class FFN:
    def __init__(self , input_size , output_size , layers_sizes , nlayers , random_state=None):
        
        self.weights = []
        self.biases = []
        self.activations = []
        self.nlayers = nlayers
        self.inp = input_size
        self.out = output_size
        self.layer_sizes = layers_sizes
        self.score = 0
        
        if random_state:
            np.random.seed(random_state)

        for i in range(nlayers):
            if i == 0:
                # The first layer weight matrix includes the input size
                self.weights.append(np.random.randn(input_size, layers_sizes))
                self.biases.append(np.random.randn(layers_sizes))
                self.activations.append(np.random.choice(range(8), layers_sizes))
            elif i == nlayers - 1:
                # The last layer weight matrix includes the output size
                self.weights.append(np.random.randn(layers_sizes, output_size))
                self.biases.append(np.random.randn(output_size))
                self.activations.append(np.random.choice(range(8), output_size))
            else:
                # Subsequent layers have weights only based on the layer size
                self.weights.append(np.random.randn(layers_sizes, layers_sizes))
                self.biases.append(np.random.randn(layers_sizes))
                self.activations.append(np.random.choice(range(8), layers_sizes))



    def forward(self , x):
        x = np.array(x)
        for i in range(self.nlayers):
            x = np.dot(x, self.weights[i]) + self.biases[i]
            x = self.activation(x , self.activations[i])
        return x

    def activation(self, x, i):
        farr = [binstep, linear, sigmoid, tanh, relu, lrelu, elu, swish]
        return np.vectorize(lambda a, b: farr[b](a))(x, i)
    
    def cust_set(self,weights,biases,activations):
        # When we want to set the properties externally (Hoping that the sizes of the net wouldn't change)
        self.weights = weights
        self.biases = biases
        self.activations = activations 




import numpy as np
from NN import FFN
class Population:
    def __init__(self ,  input_size , output_size , layers_sizes , nlayers , np_nr, population_size, parent_percentage , mutation_rate , random_state=None):
        self.size = population_size
        self.mut_rate = mutation_rate
        self.population = [FFN(input_size = input_size,output_size=output_size,layers_sizes=layers_sizes,nlayers=nlayers) for _ in range(population_size)]
        self.npnr = np_nr
        # children to random ratio
        self.pr = parent_percentage
    def get_populations(self):
        return self.population
    
    def rank(self,Pop):
        scores = np.array([ffn.score for ffn in Pop])
        sorted_indices = np.argsort(scores)[::-1]  # reverse order to get descending
        sorted_ffns = np.array(Pop)[sorted_indices]  # use the indices to sort the array
        return sorted_ffns
        
    def select_parents(self):
        # select the best pr(parent rate) of NN layers as parents
        num_parents = int(self.size * self.pr)
        parents = self.population[:num_parents]
        # flatten the parents array to 1D
        parents = np.array(parents).flatten()
        return parents

    def breed_population(self, parents):
        # create new population by mixing parents' weights and biases with some mutation
        new_population = []
        num_children = int(self.size * self.npnr)
        num_random = self.size - num_children
        parentref = parents[0]
        # create children from parents
        for i in range(num_children):
            parent1, parent2 = np.random.choice(parents, size=2, replace=False)
            child_weight = []
            child_bias = []
            child_act = []
        for j, (p1_weight, p2_weight) in enumerate(zip(parent1.weights, parent2.weights)):
            split_point = np.random.randint(0, len(p1_weight))
            new_weight = np.concatenate((p1_weight[:split_point], p2_weight[split_point:]))
            # add noise to the child weights based on the mutation rate
            if np.random.rand() < self.mut_rate:
                new_weight += np.random.normal(scale=0.1, size=new_weight.shape)
            child_weight.append(new_weight)

        for j, (p1_bias, p2_bias) in enumerate(zip(parent1.biases, parent2.biases)):
            split_point = np.random.randint(0, len(p1_bias))
            new_bias = np.concatenate((p1_bias[:split_point], p2_bias[split_point:]))
            # add noise to the child biases based on the mutation rate
            if np.random.rand() < self.mut_rate:
                new_bias += np.random.normal(scale=0.1, size=new_bias.shape)
            child_bias.append(new_bias)

        for j, (p1_act, p2_act) in enumerate(zip(parent1.activations, parent2.activations)):
            split_point = np.random.randint(0, len(p1_act))
            new_act = np.concatenate((p1_act[:split_point], p2_act[split_point:]))
            child_act.append(new_act)
            child = FFN(parent1.inp, parent1.out, parent1.layer_sizes, parent1.nlayers)
            child.cust_set(child_weight,child_bias,child_act)
            new_population.append(child)

        # create random mutations
        for i in range(num_random):
            child = FFN(parentref.inp, parentref.out, parentref.layer_sizes, parentref.nlayers)
            new_population.append(child)
        self.population = new_population    
    def evolve(self, RatedPopN):
        self.population = self.rank(RatedPopN)
        parents = self.select_parents()
        self.breed_population(parents)
        return self.population
    def getbest(self):
        self.population = self.rank(self.population)
        return self.population[0]


from NNevo import Population
import NN
import matplotlib.pyplot as plt
input_set = [[0,0],[0,1],[1,0],[1,1]]
output_set = [0,1,1,1]
Pop = Population(input_size = 2, output_size = 1 , layers_sizes = 2 , nlayers = 3, np_nr = 0.70 ,population_size = 500,parent_percentage = 0.10,mutation_rate = 0.2)

ourpop = Pop.get_populations()
aths = 0
athss = None
hsgen=[]
for i in range(500):
    # running for about 500 generations
    hs = 0
    for j in range(len(ourpop)):
        score = 4.0
        for k in range(len(input_set)):
            res = ourpop[j].forward(input_set[k])
            score -=  (res - output_set[k])**2
        ourpop[j].score = score
        if score>hs:
            hs = score
        if score > aths:
            aths = score
            athss = ourpop[j]    
    hsgen.append(hs) 
    ourpop = Pop.evolve(ourpop)
    
for i in input_set:
    print(athss.forward(i))
print(aths)
plt.plot(hsgen)
plt.show()


from setuptools import setup, find_packages

setup(
    name='SynapEvo',
    version='0.0.0',
    license='MIT',
    author="Ashutosh Adhikari,Ananya Datta,Aditya Kothari",
    author_email='email@example.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/Adhikari-Ashutosh/DARWiN',
    keywords='ML NeuralNET',
    install_requires=[
          'numpy',
      ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
)